/*
 * TimerQueue - Timer callback handler
 * Copyright (c) 2012 by Project Waffle / TSUJIKAWA Tohru <tsujikawa@tsg.ne.jp>
 *  
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

///////////////////////////////////////////////////////////////////////////////
// TimerQueue.cpp

#include	"TimerQueue.h"
#include	"Trace.h"

using namespace std;


TimerQueue*		wi_timer_q;


TimerQueueTimer::TimerQueueTimer ()
{
	timer = NULL; parent_timer_q = nullptr;
	id = 0; cat1 = cat2 = 0;
	triggered_time = expected_firing_time = duration_msec = 0;
	param = nullptr;
}


TimerQueue::TimerQueue (HWND _hwnd_message_sent_to)
{
	mutex = CREATE_MUTEX();
	timer_q = CreateTimerQueue();
	_ASSERT(timer_q);
	hwnd_message_sent_to = _hwnd_message_sent_to;
	next_id = 1;
}

TimerQueue::~TimerQueue ()
{
	if (timer_q) {
		BOOL	r = DeleteTimerQueueEx(timer_q, (HANDLE)-1);
		_ASSERT(r);
		timer_q = 0;
	}
	if (mutex) DESTROY_MUTEX(mutex);
}


void WINAPI TimerQueue::WaitOrTimerCallback (PVOID lpParameter, BOOL TimerOrWaitFired)
{
	TimerQueueTimer*		This = (TimerQueueTimer*)lpParameter;
	_ASSERT(This);
	if (!This) return;
	TimerQueue*				Parent = This->parent_timer_q;
	_ASSERT(Parent);
	if (!Parent) return;
	MUTEX_LOCK(Parent->mutex); {
		do {
			map<DWORD, pTimerQueueTimer>::const_iterator p = Parent->timers.find(This->id);
			if (p == Parent->timers.end()) break;
			if (TimerOrWaitFired) {
				PostMessage(Parent->hwnd_message_sent_to, WM_TIMERQUEUE, (WPARAM)(This->cat1), (LPARAM)(This->param));
			}
			Parent->timers.erase(p);
		} while (false);
	} MUTEX_UNLOCK(Parent->mutex);
}


DWORD TimerQueue::get_next_id ()
{
	DWORD	r = 0;
	MUTEX_LOCK(mutex); {
		while (true) {
			map<DWORD, pTimerQueueTimer>::const_iterator p = timers.find(r = next_id++);
			if (next_id == 0) next_id++;
			if (p == timers.end()) break;
		}
	} MUTEX_UNLOCK(mutex);

	return r;
}

DWORD TimerQueue::start (DWORD _cat1, DWORD _cat2, unsigned long long _duration_msec, void* _param)
{
	_ASSERT(_cat1); _ASSERT(_cat2);
	if (!_cat1 || !_cat2) return 0;

	pTimerQueueTimer	ptqt = TimerQueueTimer::new_inst();
	ptqt->parent_timer_q = this;
	ptqt->id = get_next_id(); ptqt->cat1 = _cat1; ptqt->cat2 = _cat2;
	ptqt->duration_msec = _duration_msec;
	ptqt->param = _param;

	FILETIME	ft;
	GetSystemTimeAsFileTime(&ft);
	ptqt->triggered_time = (unsigned long long(ft.dwHighDateTime) << 32) | unsigned long long(ft.dwLowDateTime);
	ptqt->expected_firing_time = ptqt->triggered_time + ptqt->duration_msec*10000;

#if 0
	AQT(("[TimerQueue::start]\n  duration_msec            = %llu\n  triggered_time           = %llu\n  expected_firing_time     = %llu\n", ptqt->duration_msec, ptqt->triggered_time, ptqt->expected_firing_time));
#endif

	BOOL	r = CreateTimerQueueTimer(&ptqt->timer, timer_q, (WAITORTIMERCALLBACK)TimerQueue::WaitOrTimerCallback, (PVOID)(ptqt.get()),
									  (DWORD)_duration_msec, 0, WT_EXECUTEINTIMERTHREAD);
	_ASSERT(r);
	if (!r) return 0;
	MUTEX_LOCK(mutex); {
		timers.insert(pair<DWORD, pTimerQueueTimer>(ptqt->id, ptqt));
	} MUTEX_UNLOCK(mutex);
	return ptqt->id;
}


bool TimerQueue::restart (DWORD _cat1, DWORD _cat2, unsigned long long new_duration_msec)
{
	FILETIME	ft;
	GetSystemTimeAsFileTime(&ft);
	unsigned long long currentTimeMillis = (unsigned long long(ft.dwHighDateTime) << 32) | unsigned long long(ft.dwLowDateTime);

	const unsigned long long GUARD_PERIOD = 1000000;	// 100ms

	MUTEX_LOCK(mutex); {
		for (map<DWORD, pTimerQueueTimer>::const_iterator p = timers.begin(); p != timers.end(); p++) {
			if ((_cat1 && _cat2 && p->second->cat1 == _cat1 && p->second->cat2 == _cat2) ||
				(_cat1 && !_cat2 && p->second->cat1 == _cat1) ||
				(!_cat1 && !_cat2)) {

				unsigned long long expected_new_firing_time = p->second->triggered_time + new_duration_msec*10000;

#if 0
				AQT(("[TimerQueue::restart]\n  new_duration_msec        = %llu\n  currentTimeMillis        = %llu\n"
					 "  duration_msec            = %llu\n  triggered_time           = %llu\n  expected_firing_time     = %llu\n  expected_new_firing_time = %llu\n",
					 new_duration_msec, currentTimeMillis,
					 p->second->duration_msec, p->second->triggered_time, p->second->expected_firing_time,
					 expected_new_firing_time));
#endif

				/* ύXO̖ƕύX̖̍K[h^C}ȉȂΉȂ */
				if ((expected_new_firing_time >= p->second->expected_firing_time && expected_new_firing_time - p->second->expected_firing_time < GUARD_PERIOD) ||
					(expected_new_firing_time <  p->second->expected_firing_time && p->second->expected_firing_time - expected_new_firing_time < GUARD_PERIOD)) {
					continue;
				} else if (currentTimeMillis + GUARD_PERIOD > expected_new_firing_time) {	/* +K[h ȓɕύX^C}\肩 */
					/* +K[h ȓɕύXO^C}\ȂΉȂ */
					if (currentTimeMillis + GUARD_PERIOD > p->second->expected_firing_time) continue;
					/* ύX^C}K[hԌƂ */
					p->second->expected_firing_time = expected_new_firing_time = currentTimeMillis + GUARD_PERIOD;
					p->second->duration_msec = new_duration_msec;
				} else {	/* +K[h ȓɕύX^C}\łȂ */
					p->second->expected_firing_time = expected_new_firing_time;
					p->second->duration_msec = new_duration_msec;
				}

				BOOL	r = ChangeTimerQueueTimer(timer_q, p->second->timer, ULONG((expected_new_firing_time - currentTimeMillis)/10000), 0);
				_ASSERT(r);
				AQT(("[TimerQueue::restart] CHANGED duration: %llu msec\n", unsigned long long(ULONG((expected_new_firing_time - currentTimeMillis)/10000))));
			}
		}
	} MUTEX_UNLOCK(mutex);

	return true;
}


bool TimerQueue::cancel (DWORD _id)
{
	_ASSERT(_id);
	HANDLE	timer = 0;
	MUTEX_LOCK(mutex); {
		map<DWORD, pTimerQueueTimer>::const_iterator p = timers.find(_id);
		if (p != timers.end()) timer = p->second->timer;
		timers.erase(p);
	} MUTEX_UNLOCK(mutex);

	if (timer) {
		BOOL	r = DeleteTimerQueueTimer(timer_q, timer, (HANDLE)-1);
		_ASSERT(r);
		if (!r) return false;
	}

	return true;
}


bool TimerQueue::cancel (DWORD _cat1, DWORD _cat2)
{
	vector<HANDLE>	timer_list;
	MUTEX_LOCK(mutex); {
		for (map<DWORD, pTimerQueueTimer>::const_iterator p = timers.begin(); p != timers.end(); ) {
			if ((_cat1 && _cat2 && p->second->cat1 == _cat1 && p->second->cat2 == _cat2) ||
				(_cat1 && !_cat2 && p->second->cat1 == _cat1) ||
				(!_cat1 && !_cat2)) {
				timer_list.push_back(p->second->timer);
				delete p->second->param; p->second->param = nullptr;
				timers.erase(p); p = timers.begin(); continue;
			}
			p++;
		}
	} MUTEX_UNLOCK(mutex);
	for (size_t i = 0; i < timer_list.size(); i++) {
		BOOL	r = DeleteTimerQueueTimer(timer_q, timer_list[i], (HANDLE)-1);
		_ASSERT(r);
		//if (!r) return false;
	}

	return true;
}
